# -*- coding: utf-8 -*-
# cython: language_level=3
# Copyright (c) 2021-present VincentRPS

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE
"""Implementation of Discord Webhooks."""

import typing
from contextvars import ContextVar

from discord import utils

from .api.rest import RESTClient, Route
from .embed import Embed
from .file import File
from .snowflake import Snowflakeish

__all__: typing.List[str] = ["Webhook", "WebhookAdapter"]


class WebhookAdapter:
    """The base class for interperting Webhooks

    .. versionadded:: 0.3.0

    Parameters
    ----------
    id
        The webhook id
    token
        The webhook token

    Attributes
    ----------
    rest
        An instance of RESTClient.
    """

    def __init__(self):
        self.rest = RESTClient()

    def fetch_webhook(self, id, token):
        """Fetch the current Webhook from the API."""
        return self.request("GET", f"/webhooks/{id}/{token}")

    def modify_webhook(
        self,
        id,
        token,
        name: typing.Optional[str] = None,
        avatar: typing.Optional[str] = None,
    ):
        """Modify the Webhook

        Parameters
        ----------
        name
            Change the name
        avatar
            Change the avatar
        """
        json = {}
        if name:
            json["name"] = name
        if avatar:
            json["avatar"] = avatar
        return self.rest.send(
            Route(
                "PATCH",
                f"/webhooks/{id}/{token}",
                webhook_id=id,
                webhook_token=token,
            ),
            json=json,
        )

    def delete_webhook(self, id, token):
        """Deletes the Webhook"""
        return self.rest.send(
            Route(
                "DELETE",
                f"/webhooks/{id}/{self.token}",
                webhook_id=id,
                webhook_token=token,
            )
        )

    def fetch_message(self, id, token, message: Snowflakeish):
        """Fetches a Webhook message."""
        return self.rest.send(
            Route(
                "GET",
                f"/webhooks/{id}/{token}/messages/{message}",
                webhook_id=id,
                webhook_token=token,
            )
        )

    def edit_message(
        self,
        id,
        token,
        message: Snowflakeish,
        content: typing.Optional[str] = None,
        allowed_mentions: typing.Optional[bool] = None,
    ):
        """Edits a Webhook message

        Parameters
        ----------
        message
            The Message ID
        content
            Change the content
        allowed_mentions
            A allowed mentions object
        """
        json = {}
        if content:
            json["content"] = content
        elif allowed_mentions:
            json["allowed_mentions"] = allowed_mentions
        return self.rest.send(
            Route(
                "POST",
                f"/webhooks/{id}/{token}/messages/{message}",
                webhook_id=id,
                webhook_token=token,
            ),
            json=json,
        )

    def delete_message(
        self,
        id,
        token,
        message: Snowflakeish,
    ):
        """Deletes a message

        Parameters
        ----------
        message
            The message to delete
        """
        return self.rest.send(
            Route(
                "DELETE",
                f"/webhooks/{id}/{token}/messages/{message}",
                webhook_id=id,
                webhook_token=token,
            )
        )

    def execute(
        self,
        id,
        token,
        content: typing.Optional[str] = None,
        username: typing.Optional[str] = None,
        avatar_url: typing.Optional[str] = None,
        tts: typing.Optional[bool] = None,
        allowed_mentions: typing.Optional[bool] = None,
        embed: typing.Optional[Embed] = None,
        embeds: typing.Optional[typing.List[Embed]] = None,
        flags: typing.Optional[typing.Any] = None,
        files: typing.Optional[typing.Sequence[File]] = None,
    ):
        """Execute the Webhook

        Parameters
        ----------
        content: :class:`str`
            The content to send.
        username: :class:`str`
            The username the Webhook should have
        avatar_url: :class:`str`
            The avatar the Webhook should have
        tts: :class:`bool`
            If the message should have tts enabled
        allowed_mentions
            A allowed mentions object
        """
        json = {}
        if content:
            json["content"] = content
        if username:
            json["username"] = username
        if avatar_url:
            json["avatar_url"] = avatar_url
        if tts:
            json["tts"] = tts
        if allowed_mentions:
            json["allowed_mentions"] = allowed_mentions
        if embed:
            if isinstance(embed, Embed):
                emb = [embed.obj]
            else:
                emb = [embed]
        if embeds:
            if isinstance(embeds, Embed):
                emb = [embed.obj for embed in embeds]
            else:
                emb = embeds
        if embed or embeds:
            json["embeds"] = emb

        if flags:
            json["flags"] = flags

        return self.rest.send(
            Route(
                "POST",
                f"/webhooks/{id}/{token}",
                webhook_id=id,
                webhook_token=token,
            ),
            json=json,
            files=files,
        )


webhook_context: ContextVar[WebhookAdapter] = ContextVar(
    "webhook_context", default=WebhookAdapter()
)


@utils.copy_doc(WebhookAdapter)
class Webhook:
    def __init__(self, id, token):
        self.id = id
        self.token = token
        self.adapter = webhook_context.get()

    @utils.copy_doc(WebhookAdapter.execute)
    def execute(
        self,
        content: typing.Optional[str] = None,
        username: typing.Optional[str] = None,
        avatar_url: typing.Optional[str] = None,
        tts: typing.Optional[bool] = None,
        allowed_mentions: typing.Optional[bool] = None,
        embed: typing.Optional[Embed] = None,
        embeds: typing.Optional[typing.List[Embed]] = None,
        flags: typing.Optional[typing.Any] = None,
        files: typing.Optional[typing.Sequence[File]] = None,
    ):
        return self.adapter.execute(
            id=self.id,
            token=self.token,
            content=content,
            username=username,
            avatar_url=avatar_url,
            tts=tts,
            allowed_mentions=allowed_mentions,
            embed=embed,
            embeds=embeds,
            flags=flags,
            files=files,
        )



# -*- coding: utf-8 -*-
# cython: language_level=3
# Copyright (c) 2021-present VincentRPS

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE


import asyncio
import random
import string
import typing
from typing import Any, Callable, Coroutine

from ..interactions.interaction import Interaction
from ..internal.dispatcher import Coro
from ..state import ConnectionState

__all__: typing.List[str] = ["Button"]


class Button:
    """Represents a Discord button interaction.

    Parameters
    ----------
    state
        The connection state
    """

    def __init__(self, state: ConnectionState):
        self.state = state

    async def _run_callback(
        self, coro: Callable[..., Coroutine[Any, Any, Any]], data, state
    ) -> None:
        try:
            await coro(Interaction(data, state))
        except asyncio.CancelledError:
            pass
        except Exception:
            raise

    async def create(
        self,
        label: str,
        callback: Coro,
        style: typing.Literal[1, 2, 3, 4, 5] = 1,
        custom_id: str = None,
        url: str = None,
    ):
        """Creates a button

        Parameters
        ----------
        label
            The button label
        callback
            The button callback
        style
            The button style to use
        custom_id
            A custom_id
        url
            A url

            .. note::

                can only be used in buttons,
                with style type 5
        """
        self.id = (
            custom_id
            if custom_id is not None
            else "".join(
                random.choice(string.ascii_letters)
                for _ in range(random.randint(10, 100))
            )
        )

        ret = {
            "type": 1,
            "components": [
                {
                    "type": 2,
                    "label": label,
                    "style": style,
                    "url": url,
                    "custom_id": self.id,
                }
            ],
        }

        self.callback = callback

        self.state.components[self] = {
            "self": self,
            "callback": self.callback,
            "data": ret,
            "id": self.id,
        }

        return ret



# -*- coding: utf-8 -*-
# cython: language_level=3
# Copyright (c) 2021-present VincentRPS

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE

from ..interactions import Interaction
from .core import Event


class OnInteraction(Event):
    """A event which processes interactions, not to be normally used

    Returns
    -------
    interaction: :class:`Interaction`
    """

    def process(self):
        for component in self.state.components.values():
            if component["id"] == self.data["data"]["custom_id"]:
                self.state.loop.create_task(
                    component["self"]._run_callback(
                        component["callback"], self.data, self.state
                    )
                )
        try:
            for application_command in self.state.application_commands.values():
                cog = application_command["cog"]
                if application_command["d"]["id"] == self.data["data"]["id"]:
                    self.state.loop.create_task(
                        application_command["self"].run(
                            application_command["callback"],
                            self.data,
                            self.state,
                            cog=cog,
                        )
                    )
        except KeyError:
            # components
            pass
        self.dispatch("INTERACTION", Interaction(self.data, self.state))



# -*- coding: utf-8 -*-
# cython: language_level=3
# Copyright (c) 2021-present VincentRPS

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE
"""Represents a Discord Embed.

ref: https://discord.dev/resources/channel#embed-limits
"""
import datetime
from typing import Any, List, Optional, Union

from discord.color import Color
from discord.colour import Colour

__all__: List[str] = ["Embed"]


class Embed:
    """Represents a Discord Embed.

    Parameters
    ----------
    title
        The embed title
    description
        The embed description
    url
        The embed url
    date
        The embed date
    color
        The embed color
    colour
        The embed colour
    timestamp
        The embed timestamp
    """

    def __init__(
        self,
        title: Optional[str] = None,
        description: Optional[str] = None,
        url: Optional[str] = None,
        date: Optional[str] = None,
        color: Optional[Union[int, Color]] = None,
        colour: Optional[Union[int, Colour]] = None,
        timestamp: datetime.datetime = None,
    ):
        self.obj: dict[str, Any] = {
            "type": "rich",
        }

        if title:
            self.obj["title"] = title
        if description:
            self.obj["description"] = description
        if url:
            self.obj["url"] = url
        if date:
            self.obj["date"] = date
        if color:
            self.obj["color"] = color
        if colour:
            self.obj["color"] = colour
        if timestamp:
            self.obj["timestamp"] = timestamp

    def to_dict(self):
        """Gives the dictionary which the embed is within

        Returns
        -------
        obj :class:`dict`
            The dict
        """
        return self.obj

    def set_footer(self, text: str = None, icon_url: str = None):
        """Sets the footer

        Parameters
        ----------
        text
            A text footer
        icon_url
            The icon url of the footer
        """
        if text:
            self.obj["footer"]["text"] = text
        if icon_url:
            self.obj["footer"]["icon_url"] = icon_url

    def remove_footer(self):
        """Removes the footer"""
        del self.obj["footer"]

    def set_thumbnail(self, url: str = None):
        """Sets a thumbnail on the embed

        Parameters
        ----------
        url
            The thumbnail url
        """
        if url is None:
            del self.obj["thumbnail"]
        else:
            self.obj["thumbnail"]["url"] = url

    def set_author(self, name: str, url: str = None, icon_url: str = None):
        """Sets the embed author

        Parameters
        ----------
        name
            The author name
        url
            The author's url
        icon_url
            The author's icon_url
        """
        self.obj["author"] = name

        if url:
            self.obj["author"]["url"] = url
        if icon_url:
            self.obj["author"]["icon_url"] = icon_url

    def remove_author(self):
        """Removes the author from the embed"""
        del self.obj["author"]

    def add_field(self, name: str, value: str, inline: bool = True):
        """Adds a field to the embed

        Parameters
        ----------
        name
            The field name
        value
            The field value
        inline
            If the field should be inline
        """
        field = {"name": name, "value": value, "inline": inline}

        try:
            self.obj["fields"].append(field)
        except KeyError:
            self.obj["fields"] = [field]

    def remove_field(self, name: str):
        """Removes the field

        Parameters
        ----------
        name
            The field to remove
        """
        del self.obj["fields"][name]



# -*- coding: utf-8 -*-
# cython: language_level=3
# Copyright (c) 2021-present VincentRPS

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE

import asyncio
from typing import Any, Callable, Coroutine, Optional

from .. import utils
from ..enums import ModalStyle
from ..interactions import Interaction
from ..state import ConnectionState


def ModalComponent(
    label: str,
    placeholder: Optional[str] = None,
    min_length: Optional[int] = 1,
    max_length: Optional[int] = 4000,
    style: ModalStyle = ModalStyle.SHORT,
    required: Optional[bool] = False,
    value: Optional[str] = None,
):
    ret = {
        "type": 1,
        "components": [
            {
                "type": 4,
                "custom_id": utils.create_snowflake(),
                "label": label,
                "style": style,
            }
        ],
    }
    if placeholder:
        ret["components"][0]["placeholder"] = placeholder
    if min_length:
        ret["components"][0]["min_length"] = min_length
    if max_length:
        ret["components"][0]["max_length"] = max_length
    if required:
        ret["components"][0]["required"] = True
    if value:
        ret["components"][0]["value"] = value

    return ret


class Modal:
    """Represents a Discord button interaction.

    Parameters
    ----------
    state
        The connection state
    """

    def __init__(self, state: ConnectionState):
        self.state = state

    async def _run_callback(
        self, coro: Callable[..., Coroutine[Any, Any, Any]], data, state
    ) -> None:
        try:
            await coro(Interaction(data, state))
        except asyncio.CancelledError:
            pass
        except Exception:
            raise

    def create(
        self,
        title: str,
        callback: Any,
        components: list[ModalComponent],
        custom_id: int = utils.create_snowflake(),
    ):
        self.id = custom_id
        ret = {"title": title, "custom_id": self.id, "components": components}
        self.state.components[self] = {
            "self": self,
            "callback": callback,
            "data": ret,
            "id": self.id,
        }
        return ret



"""
discord.ext.cogs
~~~~~~~~~~~~~~~~
Extension module to ensure the creation of Cogs.
"""
from typing import TYPE_CHECKING, Any, Callable, List, TypeVar

from ...internal import DiscordError, dispatcher

__all__: List[str] = ["Cog", "ExtensionLoadError"]

CFT = TypeVar("CFT", bound="dispatcher.CoroFunc")


class ExtensionLoadError(DiscordError):
    ...


class Cog:
    listeners: dict[str, Any] = {}
    guild_commands: dict[str, Any] = {}
    global_commands: dict[str, Any] = {}

    @property
    def __cog_name__(self) -> str:
        return type(self).__name__

    @classmethod
    def slash_command(
        cls,
        name: str = None,
        options: List[dict] = None,
        guild_ids: List[int] = None,
        default_permission: bool = True,
    ):
        """Creates a slash command

        Parameters
        ----------
        name: :class:`str`
            The slash command name
        callback
            The slash command callback
        options: :class:`List`
            A list of slash command options
        guild_ids: List[:class:`int`]
            A list of guild ids
        description: :class:`str`
            The application command description
        default_permission: :class:`bool`
            If this slash command should have default permissions
        """

        def decorator(func: CFT) -> CFT:
            name = func.__name__  # if name is None else name # fix this somehow?
            description = func.__doc__

            a = func
            if isinstance(a, staticmethod):
                a = a.__func__

            if guild_ids is not None:
                Cog.guild_commands[name] = {
                    "guild_id": guild_ids,
                    "name": name,
                    "description": description,
                    "callback": a,
                    "options": options,
                    "default_permission": default_permission,
                }
            else:
                Cog.global_commands[name] = {
                    "name": name,
                    "description": description,
                    "callback": a,
                    "options": options,
                    "default_permission": default_permission,
                }

            return func

        return decorator

    @classmethod
    def listener(cls, name: str = None) -> Callable[[CFT], CFT]:
        """Listen to a event

        Parameters
        ----------
        name
            The event to listen to
        """

        def decorator(func: CFT) -> CFT:
            a = func
            if isinstance(a, staticmethod):
                a = a.__func__
            Cog.listeners[name] = a
            return func

        return decorator

    def _inject(self, bot_self):
        self.bot = bot_self
        return self

    def _eject(self, bot_self):
        for name in self.listeners.values():
            bot_self.state.listeners.pop(name)
            self.listeners.pop(name)



# -*- coding: utf-8 -*-
# cython: language_level=3
# Copyright (c) 2021-present VincentRPS

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE
"""Represents a Discord asset, i.e a Emoji or Sticker.

ref: https://discord.dev/resources/emoji
"""
from .enums import StickerFormatType, StickerType
from .state import ConnectionState
from .types import Dict
from .user import User


class Emoji:
    """Represents a Discord Emoji.

    .. versionadded:: 0.8.0

    Parameters
    ----------
    data
        The raw emoji data

    Attributes
    ----------
    from_dict
        The raw emoji data
    """

    def __init__(self, data: Dict):
        self.from_dict = data
        """The data in dict format."""

    @property
    def id(self) -> int:
        return self.from_dict["id"]

    @property
    def name(self) -> str:
        return self.from_dict["name"]

    # def roles(self) -> Role:

    @property
    def creator(self) -> User:
        return User(self.from_dict["user"])

    def require_colons(self) -> bool:
        return self.from_dict["require_colons"]

    def managed(self) -> bool:
        return self.from_dict["managed"]

    def animated(self) -> bool:
        return self.from_dict["animated"]

    def available(self) -> bool:
        return self.from_dict["available"]


class Sticker:
    """Represents a Discord Sticker.

    .. versionadded:: 0.8.0

    Parameters
    ----------
    data: :class:`dict`
        The raw Sticker data
    state: :class:`ConnectionState`
        The connection state

    Attributes
    ----------
    from_dict
        The raw Sticker data
    """

    def __init__(self, data: dict, state: ConnectionState):
        self.from_dict = data
        self.state = state

    @property
    def id(self) -> int:
        return self.from_dict["id"]

    @property
    def pack(self) -> int:
        return self.from_dict["pack_id"]

    @property
    def name(self) -> str:
        return self.from_dict["name"]

    def description(self) -> str:
        return self.from_dict["description"]

    def tags(self) -> str:
        return self.from_dict["tags"]

    def type(self) -> StickerType:
        if self.from_dict["type"] == 1:
            return StickerType.STANDARD
        else:
            return StickerType.GUILD

    def format(self) -> StickerFormatType:
        if self.from_dict["format_type"] == 1:
            return StickerFormatType.PNG
        elif self.from_dict["format_type"] == 2:
            return StickerFormatType.APNG
        else:
            return StickerFormatType.LOTTIE

    def available(self) -> bool:
        return self.from_dict["available"]

    def guild_id(self) -> int:
        return int(self.from_dict["guild_id"])

    def creator(self) -> User:
        return User(self.from_dict["user"])

    def sort_value(self) -> int:
        return int(self.from_dict["sort_value"])


class Attachment:
    """Represents a Discord Attachment

    .. versionadded:: 0.8.0

    Parameters
    ----------
    data: :class:`dict`
        The raw attachment data
    """

    def __init__(self, data: dict):
        self.from_dict = data

    @property
    def id(self) -> int:
        """Gives the attachment' id

        Returns
        -------
        :class:`int`
        """
        return self.from_dict["id"]

    @property
    def name(self) -> str:
        """Gives the attachments name

        Returns
        -------
        :class:`str`
        """
        return self.from_dict["filename"]

    @property
    def description(self) -> str:
        """Gives the attachments description

        Returns
        -------
        :class:`str`
        """
        return self.from_dict["description"]

    def type(self) -> str:
        """Gives the content type, i.e 'image/png' or 'image/jpeg'

        Returns
        -------
        :class:`str`
        """
        return self.from_dict["content_type"]

    def size(self) -> int:
        """Gives the size of the attachment

        Returns
        -------
        :class:`int`
        """
        return self.from_dict["size"]

    def url(self) -> str:
        """Gives the attachment' url

        Returns
        -------
        :class:`str`
        """
        return self.from_dict["url"]

    def proxy_url(self) -> str:
        """Gives the proxied url of the attachment

        Returns
        -------
        :class:`str`
        """
        return self.from_dict["proxy_url"]

    def height(self) -> int:
        """Gives the height of the attachment

        Returns
        -------
        :class:`int`
        """
        return self.from_dict["height"]

    def width(self) -> int:
        """Gives the width of the attachment

        Returns
        -------
        :class:`int`
        """
        return self.from_dict["width"]

    def invisable(self) -> bool:
        """If the attachment is normally seeable or not

        Returns
        -------
        :class:`bool`
        """
        return self.from_dict["ephemeral"]



# -*- coding: utf-8 -*-
# cython: language_level=3
# Copyright (c) 2021-present VincentRPS

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE
"""Represents a Discord User

ref: https://discord.dev/resources/users
"""
from typing import List

__all__: List[str] = ["User"]


class User:
    """Represents a Discord User

    .. versionadded:: 0.6.0

    Parameters
    ----------
    usr
        The user in dict format
    """

    def __init__(self, usr: dict):
        self.from_dict = usr

    def username(self) -> str:
        """The users username

        Returns
        -------
        :class:`str`
        """
        return self.from_dict["username"]

    def discriminator(self) -> int:
        """The users discriminator id

        Returns
        -------
        :class:`int`
        """
        return self.from_dict["discriminator"]

    @property
    def id(self) -> int:
        """The users snowflake id

        Returns
        -------
        :class:`int`
        """
        return self.from_dict["id"]

    def public_flags(self):
        """The users public flags."""
        return self.from_dict["public_flags"]

    @property
    def bot(self) -> bool:
        """If the User is a bot or not

        Returns
        -------
        :class:`bool`
        """
        return self.from_dict["bot"]



# -*- coding: utf-8 -*-
# cython: language_level=3
# Copyright (c) 2021-present VincentRPS

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE
from __future__ import annotations

import asyncio
import json
import logging
import typing
import weakref
from urllib.parse import quote

import aiohttp

from discord import utils
from discord.file import File
from discord.internal.exceptions import Forbidden, NotFound, RESTError, ServerError
from discord.state import ConnectionState
from discord.types.dict import Dict

_log = logging.getLogger(__name__)

__all__: typing.List[str] = [
    "RESTClient",
]

PAD = typing.TypeVar("PAD", bound="PadLock")

aiohttp.hdrs.WEBSOCKET = "websocket"


async def parse_tj(
    response: aiohttp.ClientResponse,
) -> typing.Union[Dict, str]:
    text = await response.text(encoding="utf-8")
    try:
        if response.headers["content-type"] == "application/json":
            return json.loads(text)  # type: ignore
    except KeyError:
        # could be errored out
        # cause of cloudflare
        pass

    return text


class Route:
    def __init__(self, method: str, endpoint: str, **params: typing.Any):
        self.method = method
        self.endpoint = endpoint
        self.url = "https://discord.com/api/v9" + endpoint

        self.guild_id: typing.Optional[int] = params.get("guild_id")
        self.channel_id: typing.Optional[int] = params.get("channel_id")

        # Webhooks
        self.webhook_id: typing.Optional[int] = params.get("webhook_id")
        self.webhook_token: typing.Optional[str] = params.get("webhook_token")

    @property
    def bucket(self) -> str:
        return f"{self.method}:{self.endpoint}:{self.guild_id}:{self.channel_id}:{self.webhook_id}:{self.webhook_token}"  # type: ignore # noqa: ignore


class PadLock:
    # based off the PadLock of interactions.py & MaybeUnlock of discord.py.
    def __init__(self, lock: asyncio.Lock):
        self.lock: asyncio.Lock = lock
        self.MaybeUnlock: bool = True

    def __enter__(self: PAD) -> PAD:
        return self

    # defers the UnLock.
    def defer(self) -> None:
        self.MaybeUnlock = False

    def __exit__(self, exc_type, exc, traceback) -> None:
        if self.MaybeUnlock:
            self.lock.release()


class RESTClient:
    """Represents a Rest connection with Discord.

    .. versionadded:: 0.3.0

    Attributes
    ----------
    url
        The Discord API URL
    loop
        The current event loop or your own.
    connector
        The base aiohttp connector
    header
        The header sent to discord.
    """

    def __init__(self, *, state=None, proxy=None, proxy_auth=None):
        self.user_agent = "DiscordBot https://github.com/VincentRPS/discord.io"
        self.header: typing.Dict[str, str] = {"User-Agent": self.user_agent}
        self._locks: weakref.WeakValueDictionary[
            str, asyncio.Lock
        ] = weakref.WeakValueDictionary()
        self._has_global: asyncio.Event = asyncio.Event()
        self._has_global.set()
        self.state = state or ConnectionState()
        self.proxy = proxy
        self.proxy_auth = proxy_auth
        self._session: aiohttp.ClientSession = utils.MISSING

    async def send(  # noqa: ignore
        self,
        route: Route,
        files: typing.Optional[typing.Sequence[File]] = None,
        form: typing.Optional[typing.Iterable[Dict]] = None,
        **params: typing.Any,
    ):
        """Sends a request to discord

        .. versionadded:: 0.3.0
        """
        self._session = aiohttp.ClientSession()
        method = route.method
        url = route.url
        bucket = route.bucket

        lock = self._locks.get(bucket)

        if lock is None:
            lock = asyncio.Lock()
            if bucket is not None:
                self._locks[bucket] = lock

        if self.proxy is not None:
            params["proxy"] = self.proxy

        if self.proxy_auth is not None:
            params["proxy_auth"] = self.proxy_auth

        if "json" in params:
            self.header["Content-Type"] = "application/json"  # Only json.
            params["data"] = json.dumps(params.pop("json"))

        if "token" in params:
            self.header["Authorization"] = "Bot " + params.pop("token")

        if "reason" in params:
            self.header["X-Audit-Log-Reason"] = quote(params.pop("reason"), "/ ")

        params["headers"] = self.header

        if not self._has_global.is_set():
            await self._has_global.wait()

        await lock.acquire()
        with PadLock(lock) as padl:
            for tries in range(5):

                if files:
                    for f in files:
                        f.reset(seek=tries)

                if form:
                    form_data = aiohttp.FormData()
                    for kwargs in form:
                        form_data.add_field(**kwargs)

                    params["data"] = form_data

                try:
                    async with self._session.request(method, url, **params) as r:

                        d = await parse_tj(r)
                        _log.debug(
                            "< %s %s %s %s", method, url, params.get("data"), bucket
                        )

                        try:
                            remains = r.headers.get("X-RateLimit-Remaining")
                            reset_after = r.headers.get("X-RateLimit-Reset-After")
                        except KeyError:
                            # Some endpoints don't give you these ratelimit headers
                            # and so they will error out.
                            pass

                        if remains == "0" and r.status != 429:
                            # the bucket was depleted
                            padl.defer()
                            _log.debug(
                                "A ratelimit Bucket was depleted. (bucket: %s, retry: %s)",
                                bucket,
                                float(reset_after),
                            )
                            self.state.loop.call_later(float(reset_after), lock.release)

                        if r.status == 429:
                            if not r.headers.get("via") or isinstance(d, str):
                                # handles couldflare bans
                                raise RESTError(d)

                            retry_in: float = d["retry_after"]
                            _log.warning(
                                "The Rest Client seems to be ratelimited,"
                                " Retrying in: %.2f seconds. Handled with the bucket: %s",
                                retry_in,
                                bucket,
                            )

                            is_global = d.get("global", False)

                            if is_global:
                                _log.debug(
                                    "Global ratelimit was hit, retrying in %s", retry_in
                                )
                                self._has_global.clear()

                            await asyncio.sleep(retry_in)
                            _log.debug(
                                "Finished waiting for the ratelimit, now retrying..."
                            )

                            if not is_global:
                                self._has_global.set()
                                _log.debug("Global ratelimit has been depleted.")

                            continue

                        elif r.status == 403:
                            raise Forbidden(d)
                        elif r.status == 404:
                            raise NotFound(d)
                        elif r.status == 500:
                            raise ServerError(d)
                        elif 300 > r.status >= 200:
                            _log.debug("> %s", d)
                            await self.close()
                            return d
                        elif r.status == 204:
                            pass
                        else:
                            _log.error(r)

                except Exception as exc:
                    raise RESTError(f"{exc}")

    async def cdn(self, url) -> bytes:
        async with self._session.get(url) as r:
            if r.status == 200:
                return await r.read()
            elif r.status == 404:
                raise NotFound(f"{url} is not found!")
            elif r.status == 403:
                raise Forbidden("You are not allowed to see this image!")
            else:
                raise RESTError(r, "asset recovery failed.")

    async def close(self) -> None:
        if self._session:
            await self._session.close()  # Closes the session

    async def create_if_not_exists(self):
        if self._session is None or utils.MISSING:
            self._session = aiohttp.ClientSession()
        else:
            pass



# -*- coding: utf-8 -*-
# cython: language_level=3
# Copyright (c) 2021-present VincentRPS

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE


from ..channels import StageInstance, TextChannel
from ..guild import Guild
from ..member import Member
from ..user import User
from .core import Event


class OnStageInstanceCreate(Event):
    """Called when a stage instance is created

    Returns
    -------
    Stage: :class:`StageInstance`
    """

    def process(self):
        ret = StageInstance(self.data)
        self.state.stage_instances.new(self.data["id"], self.data)

        self.dispatch("stage_instance_create", ret)


class OnStageInstanceEdit(Event):
    """Gets called when a stage instance is edited

    Returns
    -------
    before: :class:`StageInstance`, can be :class:`None`
    after: :class:`StageInstance`
    """

    def process(self):
        before = StageInstance(self.state.stage_instances.get(self.data["id"]))
        after = StageInstance(self.data)
        self.state.stage_instances.edit()

        self.dispatch("stage_instance_edit", before, after)


class OnStageInstanceDelete(Event):
    """Called when a stage instance is deleted

    Returns
    -------
    Stage: :class:`StageInstance`
    """

    def process(self):
        ret = StageInstance(self.data)
        self.state.stage_instances.pop(self.data["id"])

        self.dispatch("stage_instance_delete", ret)


class OnTyping(Event):
    """Gets called when a user starts typing

    Returns
    -------
    Member: :class:`Member`
    User: :class:`User`
    Channel: :class:`TextChannel`
    timestamp: :class:`str`
    """

    def process(self):
        channel = TextChannel(
            self.state.channels.get(self.data["channel_id"]), self.state
        )
        guild = Guild(
            self.state.guilds.get(self.data["guild_id"]), self.state.app.factory
        )
        user = User(self.state.members.get(self.data["user_id"]))
        timestamp: int = self.data["timestamp"]
        member = Member(self.data["member"], guild.id, self.state.app.factory)

        self.dispatch("typing", member, user, channel, timestamp)


class OnInviteCreate(Event):
    """Called when a invite is created

    Returns
    -------
    Channel: :class:`TextChannel`
    Code: :class:`str`
    created_at: :class:`str`
    guild: :class:`Guild`
    inviter: :class:`User`
    max_age: :class:`int`
    max_uses: :class:`int`
    target_type: :class:`int`
    target_user: :class:`User`
    target_application: :class:`Any`
    tempoary: :class:`bool`
    uses: :class:`int`
    """

    def process(self):
        channel = TextChannel(
            self.state.channels.get(self.data.get("channel_id")), self.state
        )
        code: str = self.data.get("code")
        created_at: str = self.data.get("created_at")
        guild = Guild(self.state.guilds.get(self.data["guild_id"]))
        inviter = User(self.data["inviter"])
        max_age: int = self.data.get("max_age")
        max_uses: int = self.data.get("max_uses")
        target_type: int = self.data.get("target_type")
        target_user = User(self.data.get("target_user"))
        target_application = self.data.get("target_application")
        tempoary: bool = self.data.get("tempoary")
        uses: int = self.data.get("uses")

        self.dispatch(
            "invite_create",
            channel,
            code,
            created_at,
            guild,
            inviter,
            max_age,
            max_uses,
            target_type,
            target_user,
            target_application,
            tempoary,
            uses,
        )


class OnInviteDelete(Event):
    """Gets called when a invite is deleted

    Returns
    -------
    Channel: :class:`TextChannel`
    Guild: :class:`Guild`
    Code: :class:`str`
    """

    def process(self):
        channel = TextChannel(
            self.state.channels.get(self.data.get("channel_id")), self.state
        )
        guild = Guild(
            self.state.guilds.get(self.data["guild_id"]), self.state.app.factory
        )
        code: str = self.data.get("code")

        self.dispatch("invite_delete", channel, guild, code)


class OnUserUpdate(Event):
    """Gets called when a user is updated

    Returns
    -------
    User: :class:`User`
    """

    def process(self):
        user = User(self.data)

        self.dispatch("user_update", user)


class OnWebhooksUpdate(Event):
    """Gets called when webhooks are updated

    Returns
    -------
    Guild: :class:`Guild`
    Channel: :class:`TextChannel`
    """

    def process(self):
        guild = Guild(
            self.state.guilds.get(self.data["guild_id"]), self.state.app.factory
        )
        channel = TextChannel(
            self.state.channels.get(self.data["channel_id"]), self.state
        )

        self.dispatch("webhooks_update", guild, channel)



# -*- coding: utf-8 -*-
# cython: language_level=3
# Copyright (c) 2021-present VincentRPS

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE
"""
User interfacing, made to make the library look. Cool.
based off hikari's use of colorlog.
"""

import datetime
import importlib.resources
import logging
import logging.config
import platform
import string
import sys
import time
import warnings
from typing import Optional, Union

import colorlog

from discord import __copyright__, __git_sha1__, __license__, __version__
from discord.types.dict import Dict


def start_logging(flavor: Union[None, int, str, Dict], debug: bool = False):

    if len(logging.root.handlers) != 0:
        return  # the user is most likely using logging.basicConfig or another alt.

    if flavor is None:
        if not debug:
            flavor = logging.INFO
        else:
            flavor = logging.DEBUG

    if isinstance(flavor, dict):
        logging.config.dictConfig(flavor)

        if flavor.get("handler"):
            return

        flavor = None

    # things that never will be logged.
    logging.logThreads = None
    logging.logProcesses = None

    colorlog.basicConfig(
        level=flavor,
        format="%(log_color)s%(bold)s%(levelname)-1.1s%(thin)s %(asctime)23.23s %(bold)s%(name)s: "
        "%(thin)s%(message)s%(reset)s",
        stream=sys.stderr,
        log_colors={
            "DEBUG": "cyan",
            "INFO": "green",
            "WARNING": "yellow",
            "ERROR": "red",
            "CRITICAL": "red, bg_white",
        },
    )
    warnings.simplefilter("always", DeprecationWarning)
    logging.captureWarnings(True)


def print_banner(module: Optional[str] = "discord"):
    banner = importlib.resources.read_text(module, "banner.txt")
    today = datetime.date.today()

    args = {
        "copyright": __copyright__,
        "version": __version__,
        "license": __license__,
        "current_time": today.strftime("%B %d, %Y"),
        "py_version": platform.python_version(),
        "git_sha": __git_sha1__[:8],
    }
    args.update(colorlog.escape_codes.escape_codes)

    sys.stdout.write(string.Template(banner).safe_substitute(args))
    sys.stdout.flush()
    time.sleep(0.162)  # sleep for a bit.



# -*- coding: utf-8 -*-
# cython: language_level=3
# Copyright (c) 2021-present VincentRPS

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE

import asyncio
import datetime
import inspect
import random
import time
from typing import Any, Callable, Generic, Optional, TypeVar, overload

T = TypeVar("T")

# probably should be stored in snowflake.py?
Epoch = 1420070400000

# probably missing some keys here?
class _Missing:
    @overload
    def __str__() -> str:
        ...

    @overload
    def __str__(self):
        ...

    @overload
    def __str__():
        return None

    @overload
    def __int__() -> int:
        ...

    @overload
    def __int__(self):
        ...

    @overload
    def __int__(self):
        return None

    def __bool__(self):
        return False

    def __repr__(self):
        return "..."

    def __eq__(self):
        return False


MISSING: Any = _Missing()

# i am too lazy to just copy/paste the docstrings, this is a better way
def copy_doc(original: Callable) -> Callable[[T], T]:
    def decorator(overridden: T) -> T:
        overridden.__doc__ = original.__doc__
        try:
            overridden.__signature__ = inspect.signature(original)
        except TypeError:
            pass
        return overridden

    return decorator


def utcnow():
    """Gives the current time in utc

    Returns
    -------
    :class:`int`
    """
    return datetime.datetime.now(datetime.timezone.utc)


def create_snowflake(time: Optional[datetime.datetime] = None) -> int:
    """Creates a Discord snowflake via the epoch and some simple math

    Parameters
    ----------
    time: :class:`datetime.datetime`
        The time this snowflake should've been created at,
        defaults to the current time.

    Returns
    -------
    :class:`int`
    """
    time = time or utcnow()
    return int(time.timestamp() * 1000 - Epoch) << 22 | 0x3FFFFF


async def getch(fetch, get):
    try:
        await fetch
    except KeyError:
        await get


def img_mime_type(data: bytes):
    if data.startswith(b"\x89\x50\x4E\x47\x0D\x0A\x1A\x0A"):
        return "image/png"
    elif data[0:3] == b"\xff\xd8\xff" or data[6:10] in (b"JFIF", b"Exif"):
        return "image/jpeg"
    elif data.startswith((b"\x47\x49\x46\x38\x37\x61", b"\x47\x49\x46\x38\x39\x61")):
        return "image/gif"
    elif data.startswith(b"RIFF") and data[8:12] == b"WEBP":
        return "image/webp"
    else:
        raise TypeError("Image given is not supported")


async def wait_for(futures, *, timeout):
    ensured = [asyncio.ensure_future(fut) for fut in futures]
    done, pending = await asyncio.wait(
        ensured, timeout=timeout, return_when=asyncio.ALL_COMPLETED
    )

    if len(pending) != 0:
        raise asyncio.TimeoutError()

    return done


class ExponentialBackoff(Generic[T]):
    # an exponantial backoff api based off how discord.py implemented theres.
    def __init__(self, base: int = 1, *, integral: T = False):
        self.base = base
        self.exp = 0
        self.max = 10
        self.reset_time = base * 2 ** 11
        self.last_invoke = time.monotonic()
        rand = random.Random()
        rand.seed()
        self.rand = rand.randrange() if integral else rand.uniform

    def delay(self):
        invoke = time.monotonic
        interval = invoke - self.last_invoke

        if interval > self.reset_time:
            self.exp = 0

        self.exp = min(self.exp + 1, self.max)
        return self.rand(0, self.base * 2 ** self.exp)



# -*- coding: utf-8 -*-
# cython: language_level=3
# Copyright (c) 2021-present VincentRPS

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE

import datetime
import typing
from json import dumps
from typing import Any, Optional

import aiohttp

from .. import utils
from ..enums import ScheduledEventType
from ..file import File
from ..flags import MessageFlags
from ..snowflake import Snowflakeish
from ..state import ConnectionState
from ..types import allowed_mentions
from ..types.dict import Dict
from .rest import RESTClient, Route

__all__: typing.List[str] = [
    "RESTFactory",
]


class RESTFactory:
    """The RESTFactory for most requests.

    .. versionadded:: 0.3.0

    Attributes
    ----------
    rest
        The RESTClient.
    """

    def __init__(
        self,
        state: typing.Optional[ConnectionState] = None,
        proxy: typing.Optional[str] = None,
        proxy_auth: typing.Optional[str] = None,
    ):
        self.proxy = proxy
        self.proxy_auth = proxy_auth
        self.state = state or ConnectionState()
        self.rest = RESTClient(state=self.state, proxy=proxy, proxy_auth=proxy_auth)

    async def login(
        self, token: str
    ) -> typing.Coroutine[typing.Any, typing.Any, typing.Union[typing.Any, None]]:
        self.token = token

        if len(self.token) != 59:
            raise Exception("Invalid Bot Token Was Passed")
        else:
            pass

        r = await self.rest.send(Route("GET", "/users/@me"), token=self.token)

        self.state.bot_info[self.state.app] = r

        return r

    def logout(
        self,
    ) -> typing.Coroutine[typing.Any, typing.Any, typing.Union[typing.Any, None]]:
        return self.rest.send(
            Route("POST", "/auth/logout")
        )  # Log's you out of the bot.

    def get_gateway_bot(
        self,
    ) -> typing.Coroutine[typing.Any, typing.Any, typing.Union[typing.Any, None]]:
        return self.rest.send(Route("GET", "/gateway/bot"))

    def create_message(
        self,
        channel: Snowflakeish,
        content: typing.Optional[str] = None,
        files: typing.Optional[typing.Sequence[File]] = None,
        tts: typing.Optional[bool] = False,
        embeds: typing.List[Dict] = None,
        allowed_mentions: typing.Optional[allowed_mentions.MentionObject] = None,
        message_reference: typing.Optional[dict] = None,
        components: typing.Optional[list[Dict]] = None,
    ) -> typing.Coroutine[typing.Any, typing.Any, typing.Union[typing.Any, None]]:
        json = {
            "tts": tts,
            "allowed_mentions": allowed_mentions,
        }
        if content is not None:
            json["content"] = content
        if message_reference is not None:
            json["message_reference"] = message_reference
        if components is not None:
            json["components"] = components
        if embeds is not None:
            json["embeds"] = embeds

        if files:
            form = []
            form.append({"name": "payload_json", "value": dumps(json)})
            if len(files) == 1:
                file = files[0]
                form.append(
                    {
                        "name": "file",
                        "value": file.fp,
                        "filename": file.filename,
                        "content_type": "application/octet-stream",
                    }
                )
            else:
                for index, file in enumerate(files):
                    form.append(
                        {
                            "name": f"file{index}",
                            "value": file.fp,
                            "filename": file.filename,
                            "content_type": "application/octet-stream",
                        }
                    )

            return self.rest.send(
                Route("POST", f"/channels/{channel}/messages", channel_id=channel),
                json=json,
                form=form,
                files=files,
            )

        return self.rest.send(
            Route("POST", f"/channels/{channel}/messages", channel_id=channel),
            json=json,
        )

    def get_channel(self, channel: typing.Optional[Snowflakeish] = None):
        return self.rest.send(Route("GET", f"/channels/{channel}"))

    def edit_channel(
        self,
        name: typing.Optional[str] = None,
        channel: typing.Optional[Snowflakeish] = None,
        type: typing.Optional[str] = None,
    ):
        if type == "group_dm":
            payload = {}
            if name:
                payload["name"] = name
        return self.rest.send(Route("PATCH", f"/channels/{channel}"))

    def create_invite(
        self,
        *,
        channel_id: typing.Optional[Snowflakeish] = None,
        reason: typing.Optional[str] = None,
        max_age: typing.Optional[int] = 0,
        max_uses: typing.Optional[int] = 0,
        tempoary: typing.Optional[bool] = False,
        unique: typing.Optional[bool] = True,
    ):
        json = {
            "max_age": max_age,
            "max_uses": max_uses,
            "tempoary": tempoary,
            "unique": unique,
        }
        if channel_id:
            json["channel_id"] = channel_id
        return self.rest.send(Route("POST"), reason=reason, json=json)

    # Application Commands

    # global

    def create_global_application_command(
        self,
        application_id: Snowflakeish,
        name: str,
        description: str,
        options: typing.Optional[typing.List[Dict]] = None,
        default_permission: typing.Optional[bool] = True,
        type: int = 1,
    ):
        json = {
            "name": name,
            "description": description,
            "type": type,
        }
        if default_permission is False:
            json["default_permission"] = False
        if options:
            json["options"] = options
        return self.rest.send(
            Route("POST", f"/applications/{application_id}/commands"), json=json
        )

    def get_global_application_command(
        self, application_id: Snowflakeish, command: Snowflakeish
    ):
        return self.rest.send(
            Route("GET", f"/applications/{application_id}/commands/{command}")
        )

    def edit_global_application_command(
        self,
        application_id: Snowflakeish,
        command_id: Snowflakeish,
        name: str,
        description: str,
        options: typing.Optional[typing.List[Dict]] = None,
        default_permission: typing.Optional[bool] = True,
    ):
        json = {
            "name": name,
            "description": description,
            "default_permission": default_permission,
        }
        if options:
            json["options"] = options
        return self.rest.send(
            Route("PATCH", f"/applications/{application_id}/commands/{command_id}"),
            json=json,
        )

    def get_global_application_commands(self, application_id: Snowflakeish):
        return self.rest.send(Route("GET", f"/applications/{application_id}/commands"))

    def delete_global_application_command(
        self,
        application_id: int,
        command: int,
    ):
        return self.rest.send(
            Route(
                "DELETE",
                f"/applications/{application_id}/commands/{command}",
            )
        )

    # Guild Commands

    def create_guild_application_command(
        self,
        application_id: Snowflakeish,
        guild_id: Snowflakeish,
        name: str,
        description: str,
        options: typing.Optional[typing.List[Dict]],
        default_permission: typing.Optional[bool] = True,
        type: int = 1,
    ):
        json = {
            "name": name,
            "description": description,
            "default_permission": default_permission,
            "type": type,
        }
        if default_permission is False:
            json["default_permission"] = False
        if options:
            json["options"] = options
        return self.rest.send(
            Route("POST", f"/applications/{application_id}/guilds/{guild_id}/commands"),
            json=json,
        )

    def get_guild_application_command(
        self,
        application_id: Snowflakeish,
        guild_id: Snowflakeish,
        command: Snowflakeish,
    ):
        return self.rest.send(
            Route(
                "GET",
                f"/applications/{application_id}/guilds/{guild_id}/commands/{command}",
            )
        )

    def get_guild_application_commands(
        self,
        application_id: Snowflakeish,
        guild_id: Snowflakeish,
    ):
        return self.rest.send(
            Route(
                "GET",
                f"/applications/{application_id}/guilds/{guild_id}/commands",
            )
        )

    def delete_guild_application_command(
        self,
        application_id: int,
        guild_id: int,
        command: int,
    ):
        return self.rest.send(
            Route(
                "DELETE",
                f"/applications/{application_id}/guilds/{guild_id}/commands/{command}",
                guild_id=guild_id,
            )
        )

    def edit_guild_application_command(
        self,
        application_id: Snowflakeish,
        command_id: Snowflakeish,
        guild_id: Snowflakeish,
        name: str,
        description: str,
        options: typing.Optional[typing.List[Dict]] = None,
        default_permission: typing.Optional[bool] = True,
    ):
        json = {
            "name": name,
            "description": description,
            "default_permission": default_permission,
        }
        if options:
            json["options"] = options
        return self.rest.send(
            Route(
                "PATCH",
                f"/applications/{application_id}/guilds/{guild_id}/commands/{command_id}",
            ),
            json=json,
        )

    # interaction response.

    def create_interaction_response(
        self,
        interaction_id: int,
        interaction_token: str,
        content: str,
        embeds: typing.Optional[typing.List[typing.Dict]] = None,
        tts: typing.Optional[bool] = False,
        allowed_mentions: typing.Optional[allowed_mentions.MentionObject] = None,
        flags: typing.Optional[MessageFlags] = None,
        components: typing.Optional[dict] = None,
    ):
        json = {
            "content": content,
        }
        if embeds is not None:
            json["embeds"] = embeds
        if tts is not False:
            json["tts"] = tts
        if allowed_mentions is not None:
            json["allowed_mentions"] = allowed_mentions
        if flags is not None:
            json["flags"] = flags
        if components is not None:
            json["components"] = components
        return self.rest.send(
            Route(
                "POST",
                f"/interactions/{interaction_id}/{interaction_token}/callback",
            ),
            json=json,
        )

    def get_initial_response(self, application_id, interaction_token):
        return self.rest.send(
            Route("GET", f"/webhooks/{application_id}/{interaction_token}/@original")
        )

    # TODO: Edit and Delete initial reponse.

    def create_followup_message(
        self,
        application_id,
        interaction_token,
        content: str,
        embeds: typing.Optional[typing.List[dict]] = None,
        allowed_mentions: typing.Optional[allowed_mentions.MentionObject] = None,
        components: typing.Optional[typing.List[Dict]] = None,
        flags: typing.Optional[MessageFlags] = None,
    ):
        json = {"content": content}
        if embeds is not None:
            json["embeds"] = embeds
        if allowed_mentions is not None:
            json["allowed_mentions"] = allowed_mentions
        if components is not None:
            json["components"] = components
        if flags is not None:
            json["flags"] = flags
        return self.rest.send(
            Route("POST", f"/webhooks/{application_id}/{interaction_token}"), json=json
        )

    def get_followup_message(
        self,
        application_id,
        interaction_token,
        message,
    ):
        return self.rest.send(
            Route(
                "GET",
                f"/webhooks/{application_id}/{interaction_token}/messages/{message}",
            ),
        )

    # TODO: Edit and Delete followup message

    def get_audit_log_entry(
        self,
        guild: Snowflakeish,
        user_id: typing.Optional[Snowflakeish] = None,
        action_type: typing.Optional[int] = None,
        before: typing.Optional[Snowflakeish] = None,
        limit: typing.Optional[int] = 50,
    ):
        ret = {"limit": limit}
        if user_id:
            ret["user_id"] = user_id
        if action_type:
            ret["action_type"] = action_type
        if before:
            ret["before"] = before
        if limit:
            ret["limit"] = limit
        return self.rest.send(
            Route(
                "GET",
                f"/guilds/{guild}/audit-logs",
                guild_id=guild,
            ),
            json=ret,
        )

    # emojis

    ...

    # guilds

    def get_guild_member(self, guild_id, user):
        return self.rest.send(
            Route("GET", f"/guilds/{guild_id}/members/{user}", guild_id=guild_id)
        )

    def get_guild_members(
        self, guild_id, limit: typing.Optional[int] = 1, after: typing.Optional[int] = 0
    ):
        ret = {"limit": limit, "after": after}
        return self.rest.send(
            Route("GET", f"/guilds/{guild_id}/members", guild_id=guild_id), json=ret
        )

    def modify_guild_member(
        self,
        guild_id: int,
        member: int,
        nick: typing.Optional[str] = None,
        roles: typing.Optional[typing.List[int]] = None,
        mute: typing.Optional[bool] = False,
        deaf: typing.Optional[bool] = False,
        channel_id: typing.Optional[int] = None,
        timeout: typing.Optional[str] = None,
        reason: typing.Optional[str] = None,
    ):
        ret = {
            "nick": nick,
            "roles": roles,
            "mute": mute,
            "deaf": deaf,
            "channel_id": channel_id,
            "communication_disabled_until": timeout,
        }
        return self.rest.send(
            Route("PATCH", f"/guilds/{guild_id}/members/{member}", guild_id=guild_id),
            reason=reason,
            json=ret,
        )

    def get_guild(self, guild_id: int):
        return self.rest.send(Route("GET", f"/guilds/{guild_id}", guild_id=guild_id))

    def modify_guild(
        self,
        guild_id: int,
        reason: Optional[str] = None,
        name: Optional[str] = None,
        verification_level: Optional[int] = None,
        default_message_notifications: Optional[int] = None,
        explicit_content_filter: Optional[int] = None,
        afk_channel_id: Optional[int] = None,
        afk_timeout: Optional[int] = None,
    ):
        ...

    # users
    def get_user(self, user: int):
        return self.rest.send(Route("GET", f"/users/{user}"))

    # scheduled events
    def get_scheduled_events(self, guild_id: int):
        return self.rest.send(
            Route("GET", f"/guilds/{guild_id}/scheduled-events", guild_id=guild_id)
        )

    def create_scheduled_event(
        self,
        guild_id: int,
        name: str,
        start_time: datetime.datetime,
        type: ScheduledEventType,
        end_time: Optional[datetime.datetime] = None,
        description: Optional[str] = None,
        privacy_level: Optional[int] = 2,
        channel_id: Optional[int] = None,
        metadata: Optional[Any] = None,
        image: Optional[File] = None,
    ):
        form = []
        json = {
            "name": name,
            "entity_type": type,
            "scheduled_start_time": start_time,
            "privacy_level": privacy_level,
        }
        if end_time:
            json["scheduled_end_time"] = end_time
        if description:
            json["description"]
        if channel_id:
            json["channel_id"] = channel_id
        if metadata:
            json["entity_metadata"] = metadata

        raw = image.fp.read(16)
        if image:
            try:
                mime = utils.img_mime_type(raw)
            except TypeError:
                mime = "application-octet-stream"
            form.append(
                {
                    "name": "image",
                    "value": image.fp,
                    "filename": image.filename,
                    "content_type": mime,
                }
            )

        return self.rest.send(
            Route("POST", f"/guilds/{guild_id}/scheduled-events", guild_id=guild_id),
            json=json,
            form=form,
            files=[image],
        )

    # assets

    def create_guild_sticker(
        self,
        guild_id: int,
        name: str,
        tags: str,
        file: File,
        reason: Optional[str] = None,
        description: Optional[str] = None,
    ):
        raw = file.fp.read(16)

        try:
            mime = utils.img_mime_type(raw)
        except TypeError:
            if raw.startswith(b"{"):
                mime = "application/json"
            else:
                mime = "application/octet-stream"

        finally:
            file.reset()

        form = [
            {
                "name": "file",
                "value": file.fp,
                "filename": file.filename,
                "content_type": mime,
            }
        ]

        if name:
            form.append({"name": "name", "value": name})
        if tags:
            form.append({"name": "tags", "value": tags})
        if description:
            form.append({"name": "description", "value": description})

        return self.rest.send(
            Route("POST", f"/guilds/{guild_id}/stickers", guild_id=guild_id),
            form=form,
            files=[file],
            reason=reason,
        )

    # if your wondering why this is here, it's because it's used in the voice gateway.
    async def ws_connect(self, url: str, *, compress: int = 0):
        kwargs = {
            "proxy_auth": self.proxy_auth,
            "proxy": self.proxy,
            "max_msg_size": 0,
            "timeout": 30.0,
            "autoclose": False,
            "headers": {
                "User-Agent": self.rest.user_agent,
            },
            "compress": compress,
        }

        sesh = aiohttp.ClientSession()
        return await sesh.ws_connect(url, **kwargs)



# MIT License

# Copyright (c) 2021 VincentRPS

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
import setuptools

__version__ = str("0.8.0")

requirements = []
with open("requirements.txt") as f:
    requirements = f.read().splitlines()

packages = [
    "discord",
    "discord.events",
    "discord.api",
    "discord.internal",
    "discord.components",
    "discord.voice",
    "discord.interactions",
    "discord.types",
    "discord.ext.cogs",
]

extra_requires = {
    "speed": [
        "orjson~=3.6.5",  # Faster alternative to the normal json module.
        "aiodns~=3.0",  # included in aiohttp speed.
        "Brotli~=1.0.9",  # included in aiohttp speed.
        "cchardet~=2.1.7",  # included in aiohttp speed.
        "ciso8601~=2.2.0",  # Faster datetime parsing.
    ],
    "voice": ["PyNaCl~=1.5.0"],
}

setuptools.setup(
    name="discord.io",
    version=__version__,
    packages=packages,
    package_data={
        "discord": ["banner.txt", "bin/*.dll"],
    },
    project_urls={
        "Documentation": "https://discordio.rtfd.io",
        "Issue Tracker": "https://github.com/VincentRPS/discord.io/issues",
        "Pull Request Tracker": "https://github.com/VincentRPS/discord.io/pulls",
    },
    url="https://github.com/VincentRPS/discord.io",
    license="MIT",
    author="VincentRPS",
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
    install_requires=requirements,
    extras_require=extra_requires,
    description="Asynchronous Discord API Wrapper For Python",
    python_requires=">=3.8",
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "License :: OSI Approved :: MIT License",
        "Intended Audience :: Developers",
        "Natural Language :: English",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: Implementation :: CPython",
        "Framework :: AsyncIO",
        "Framework :: aiohttp",
        "Topic :: Communications :: Chat",
        "Topic :: Internet :: WWW/HTTP",
        "Topic :: Internet",
        "Topic :: Software Development :: Libraries",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "Topic :: Utilities",
    ],
)



# -*- coding: utf-8 -*-
# cython: language_level=3
# Copyright (c) 2021-present VincentRPS

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE

from ..channels import TextChannel, Thread, ThreadMember, channel_parse
from ..guild import Guild
from ..member import Member
from .core import Event


class OnChannelCreate(Event):
    """Called when a channel is created

    Returns
    -------
    A channel type
    """

    def process(self):
        channel = channel_parse(self.data["type"], self.data, self.state)
        self.state.channels.new(channel.id, self.data)
        self.dispatch("channel_create", channel)


class OnChannelUpdate(Event):
    """Gets called when a channel is updated

    Returns
    -------
    A channel type
    """

    def process(self):
        before_raw = self.state.channels.get(self.data["id"])
        before = channel_parse(before_raw["type"], self.data, self.state)
        after = channel_parse(self.data["type"], self.data, self.state)
        self.dispatch("channel_edit", before, after)
        self.state.channels.edit(before.id, self.data)


class OnChannelDelete(Event):
    """Called when a channel is deleted

    Returns
    -------
    A channel type, can be :class:`None`
    """

    def process(self):
        raw = self.state.channels.get(self.data["id"])
        channel = channel_parse(raw["type"], self.data, self.state)
        self.dispatch("channel_delete", channel)
        self.state.channels.pop(self.data["id"])


class OnChannelPinsUpdate(Event):
    """Called when channel pins are updated

    Returns
    -------
    Channel: :class:`TextChannel`
    Guild: :class:`Guild`
    last_pin: :class:`str`
    """

    def process(self):
        raw = self.state.channels.get(self.data["channel_id"])
        raw_guild = self.state.guilds.get(self.data["guild_id"])
        last_pin = self.data.get("last_pin_timestamp")
        channel = TextChannel(raw, self.state)
        guild = Guild(raw_guild, self.state.app.factory)

        self.dispatch("CHANNEL_PINS_UPDATE", channel, guild, last_pin)


class OnThreadCreate(Event):
    """Called when a thread is created

    Returns
    -------
    Thread: :class:`Thread`
    """

    def process(self):
        thread = Thread(self.data, self.state)
        self.dispatch("thread_create", thread)


class OnThreadUpdate(Event):
    """Called when a thread is updated

    Returns
    -------
    before: :class:`Thread`, can be :class:`None`
    after: :class:`Thread`
    """

    def process(self):
        before_raw = self.state.channels.get(self.data["id"])
        before = Thread(before_raw)
        after = Thread(self.data, self.state)
        self.dispatch("thread_edit", before, after)
        self.state.channels.edit(before.id, self.data)


class OnThreadDelete(Event):
    """Called when a thread is deleted

    .. note::

        This event shouldn't accur on archival

    Returns
    -------
    Thread: :class:`Thread`, can be :class:`None`
    """

    def process(self):
        thread_raw = self.state.channels.get(self.data["id"])
        thread = Thread(thread_raw)
        self.dispatch("thread_delete", thread)
        self.state.channels.pop(self.data["id"])


class OnThreadListSync(Event):
    """Called when discord syncs the current threads

    Returns
    -------
    Channels: List[:class:`TextChannel`]
    threads: List[:class:`Thread`]
    members: List[:class:`Member`]
    """

    def process(self):
        guild = Guild(
            self.state.guilds.get(self.data.get("guild_id")), self.state.app.factory
        )
        channels = [
            TextChannel(channel, self.state) for channel in self.data.get("channel_ids")
        ]
        threads = [Thread(thread, self.state) for thread in self.data.get("threads")]
        members = [
            Member(member, guild.id, self.state.app.factory)
            for member in self.data.get("members")
        ]
        self.dispatch("thread_list_sync", channels, threads, members)


class OnThreadMemberUpdate(Event):
    """Called when a thread member is updated

    Returns
    -------
    Member: :class:`ThreadMember`
    Guild: :class:`Guild`
    """

    def process(self):
        member = ThreadMember(self.data)
        guild = Guild(self.data["guild_id"], self.state.app.factory)

        self.dispatch("thread_member_update", member, guild)


class OnThreadMembersUpdate(Event):
    """Called when multiple thread members are updated

    Returns
    -------
    Thread: :class:`Thread`
    Guild: :class:`Guild`
    added_members: List[:class:`ThreadMember`]
    removed_members: List[:class:`ThreadMember`]
    member_count: :class:`int`
    """

    def process(self):
        added_members = [ThreadMember(member) for member in self.data["added_members"]]
        guild = Guild(self.data["guild_id"], self.state.app.factory)
        thread = Thread(self.state.channels.get(self.data["id"]), self.state)
        member_count: int = self.data["member_count"]
        removed_members = self.data.get("removed_member_ids")

        self.dispatch(
            "thread_members_update",
            thread,
            guild,
            added_members,
            removed_members,
            member_count,
        )



# -*- coding: utf-8 -*-
# cython: language_level=3
# Copyright (c) 2021-present VincentRPS

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE
from typing import List

__all__: List[str] = [
    "DiscordWarning",
    "Deprecated",
    "NaClWarning",
]


class DiscordWarning(Warning):
    """The base warning class for discord.io"""


class Deprecated(DiscordWarning):
    """Gets raised when a object has been deprecated but not removed"""


class NaClWarning(DiscordWarning):
    """Raised when you don't have PyNaCl"""



# -*- coding: utf-8 -*-
# cython: language_level=3
# Copyright (c) 2021-present VincentRPS

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE
"""The Client powering discord.io's bots."""

import asyncio
import importlib
import importlib.machinery
import importlib.util
import logging
import sys
import time
from threading import Event
from typing import Any, Callable, List, Literal, Optional, TypeVar, Union

from discord.channels import VoiceChannel

from . import utils
from .api.gateway import Gateway
from .api.rest_factory import RESTFactory
from .components import Button, Modal, ModalComponent
from .ext.cogs import Cog, ExtensionLoadError
from .guild import Guild
from .interactions import ApplicationCommandRegistry
from .internal import dispatcher
from .state import ConnectionState
from .types.dict import Dict
from .ui import print_banner, start_logging
from .user import User
from .voice import VoiceClient, has_nacl

_log = logging.getLogger(__name__)
__all__: List[str] = ["Client"]
CFT = TypeVar("CFT", bound="dispatcher.CoroFunc")


class Client:
    """Represents a Discord bot.

    .. versionadded:: 0.4.0

    Attributes
    ----------
    factory
        The instance of RESTFactory
    state
        The client's connection state
    dispatcher
        The dispatcher
    gateway
        The Gateway
    p
        The presence
    cogs
        A :class:`dict` of all Cogs.

    Parameters
    ----------
    token
        The bot token
    intents
        The bot intents, defaults `32509`
    status
        The bot status, defaults to online
    afk
        If the bot is afk, default to False
    loop
        The loop you want to use, defaults to :class:`asyncio.new_event_loop`
    module
        The module with a `banner.txt` to print
    voice
        If to enable the voice gateway or not, defaults False.
    logs
        A :class:`int`, :class:`str` or :class:`dict`.
    debug
        To show debug logs or not.
    state: :class:`ConnectionState`
        Allow's for custom ConnectionStates,
        and soforth custom db caches.
    command_prefix: :class:`str`
        The prefix for prefixed commands,
        defaults to "".
    chunk_guild_members
        If to cache guild members,
        this allowes the before argument on member events,
        aswell as faster fetching times.
    """

    def __init__(
        self,
        loop: Optional[asyncio.AbstractEventLoop] = asyncio.new_event_loop(),
        intents: Optional[int] = 32509,
        module: Optional[str] = "discord",
        shards: Optional[int] = None,
        mobile: Optional[bool] = False,
        proxy: Optional[str] = None,
        proxy_auth: Optional[str] = None,
        logs: Optional[Union[None, int, str, Dict]] = None,
        debug: Optional[bool] = False,
        state: Optional[ConnectionState] = None,
        chunk_guild_members: Optional[bool] = False,
    ):
        print_banner(module)
        start_logging(logs, debug)
        self.state = state or ConnectionState(
            loop=loop,
            intents=intents,
            bot=self,
            shard_count=shards,
        )
        self.dispatcher = dispatcher.Dispatcher(state=self.state)
        self.factory = RESTFactory(state=self.state, proxy=proxy, proxy_auth=proxy_auth)
        self.application = ApplicationCommandRegistry(self.factory, self.state)
        self.gateway = Gateway(
            state=self.state,
            dispatcher=self.dispatcher,
            factory=self.factory,
            mobile=mobile,
        )
        self._got_gateway_bot: Event = Event()
        self.cogs: dict[str, Cog] = {}
        self._extensions = {}
        self.chunk_guild_members = chunk_guild_members

        if not has_nacl:
            _log.warning(
                "You don't have PyNaCl, meaning you won't be able to use Voice features."
            )

    async def login(self, token: str):
        """Starts the bot connection

        .. versionadded:: 0.4.0

        """
        self.token = token
        r = await self.factory.login(token)
        self.state._bot_id = r["id"]
        return r

    def voice(self, channel: VoiceChannel):
        return VoiceClient(self.state, channel)

    @property
    def latency(self) -> float:
        return self.gateway.latency

    async def connect(self, token: str):
        """Starts the WebSocket(Gateway) connection with Discord.

        .. versionadded:: 0.4.0
        """

        await self.gateway.connect(token=token)
        if self.chunk_guild_members is not False:
            # this makes sure we are already
            # connected when asking for the chunk
            await asyncio.sleep(30)
            await self.gateway._chunk_members()

    def run(self, token: str):
        """A blocking function to start your bot

        Parameters
        ----------
        token: :class:`str`
            Your bot token
        """

        async def runner():
            await self.login(token=token)
            await asyncio.sleep(0.111)  # sleep for a bit
            await self.connect(token=token)

        self.state.loop.create_task(runner())
        self.state.loop.run_forever()

    def close(self):
        self.state.loop.stop()
        self.state.loop.close()

    def fetch_guild(self, guild_id):
        """Fetches the guild from the cache

        Parameters
        ----------
        guild_id: :class:`int`
            The guild to fetch

        Returns
        -------
        :class:`Guild`
        """
        raw = self.state.guilds.get(guild_id)
        return Guild(raw, self.factory)

    def fetch_raw_guild(self, guild_id):
        return self.state.guilds.get(guild_id)

    async def get_guild(self, guild_id):
        """Gets a guild by requesting to the API

        Parameters
        ----------
        guild_id: :class:`int`
            The guild to get

        Returns
        -------
        :class:`Guild`
        """
        raw = await self.factory.get_guild(guild_id=guild_id)
        return Guild(raw, self.factory)

    async def get_voice_channel(self, channel_id: int):
        raw = await self.factory.get_channel(channel=channel_id)
        return VoiceChannel(raw, self.state)

    def create_button(
        self,
        label: str,
        callback: dispatcher.Coro,
        style: Literal[1, 2, 3, 4, 5] = 1,
        custom_id: str = None,
        url: str = None,
    ):
        """Creates a button

        Parameters
        ----------
        label
            The button label
        callback
            The button callback
        style
            The button style to use
        custom_id
            A custom_id
        url
            A url

            .. note::

                can only be used in buttons
        """
        return Button(self.state).create(label, callback, style, custom_id, url)

    @property
    def is_ready(self):
        """Returns if the bot is ready or not."""
        return self.state._ready.is_set()

    @property
    def presence(self) -> list[str]:
        return self.state._bot_presences

    def change_presence(
        self,
        name: str,
        type: int,
        status: Literal["online", "dnd", "idle", "invisible", "offline"] = "online",
        stream_url: Optional[str] = None,
        afk: Optional[bool] = False,
    ):
        """Changes the bot's presence

        Parameters
        ----------
        name
            The presence name
        type
            The presence type
        status
            The presence status

            .. note::

                can be "online", "dnd",
                invisable and offline.
        stream_url
            Used with the streaming presence type
        afk
            If to be afk or not
        """
        if type == 1 and stream_url is None:
            raise NotImplementedError("Streams need to be provided a url!")
        elif type == 1 and stream_url is not None:
            ret = {
                "name": name,
                "type": 1,
                "url": stream_url,
            }
        else:
            # another type
            ret = {
                "name": name,
                "type": type,
            }
        json = {"op": 3, "d": {"activities": [ret]}}

        if afk is True:
            json["d"]["afk"] = True
            json["d"]["since"] = time()
        else:
            json["d"]["afk"] = False
            json["d"]["since"] = None

        json["d"]["status"] = status

        return self.gateway.send(json)

    def event(self, coro: dispatcher.Coro) -> dispatcher.Coro:
        """Register an event"""
        return self.dispatcher.listen(coro)

    def listen(self, name: str = None) -> Callable[[CFT], CFT]:
        """Listen to a event

        like :meth:`Client.event` but you can have a event split into multiple coroutines.

        Parameters
        ----------
        name
            The event to listen to
        """

        def decorator(func: CFT) -> CFT:
            self.dispatcher.add_listener(func, name)
            return func

        return decorator

    @property
    def user(self):
        """Returns the bot user

        Returns
        -------
        :class:`User`
        """
        return User(self.state.bot_info[self])

    def slash_command(
        self,
        name: str = None,
        options: List[dict] = None,
        guild_ids: List[int] = None,
        default_permission: bool = True,
    ):
        """Creates a slash command

        Parameters
        ----------
        name: :class:`str`
            The slash command name
        callback
            The slash command callback
        options: :class:`List`
            A list of slash command options
        guild_ids: List[:class:`int`]
            A list of guild ids
        description: :class:`str`
            The application command description
        default_permission: :class:`bool`
            If this slash command should have default permissions
        """

        def decorator(func: CFT) -> CFT:
            name = func.__name__  # if name is None else name # fix this somehow?
            description = func.__doc__

            if guild_ids is not None:
                for guild in guild_ids:
                    self.state.loop.create_task(
                        self.application.register_guild_slash_command(
                            guild_id=guild,
                            name=name,
                            description=description,
                            callback=func,
                            options=options,
                            default_permission=default_permission,
                        )
                    )
            else:
                self.state.loop.create_task(
                    self.application.register_global_slash_command(
                        name=name,
                        description=description,
                        callback=func,
                        options=options,
                        default_permission=default_permission,
                    )
                )

            return func

        return decorator

    def add_cog(self, cog: Cog, *, override: bool = False):
        if not isinstance(cog, Cog):
            raise TypeError("ALL cog's must subclass Cog.")

        name = cog.__cog_name__
        current = self.cogs.get(name)

        if current is not None:
            if not override:
                raise TypeError("There is already another Cog with this name!")
            self.remove_cog(current)

        cog = cog._inject(self)

        for name, func in cog.listeners.items():
            self.dispatcher.add_listener(func, name, cog=cog)

        self.state.loop.create_task(self._register_cog_commands(real=cog))
        self.cogs[name] = cog

    def remove_cog(self, cog: Cog):
        self.cogs.pop(cog.__cog_name__)
        cog._eject(self)

    def _resolver(self, name: str, *, package: str):
        try:
            return importlib.util.resolve_name(name=name, package=package)
        except ImportError:
            raise TypeError("Cog is not found!")

    def _extension_loader(self, spec: importlib.machinery.ModuleSpec, key: str):
        lib = importlib.util.module_from_spec(spec)
        sys.modules[key] = lib
        try:
            spec.loader.exec_module(lib)
        except Exception as exc:
            del sys.modules[key]
            raise ExtensionLoadError(key, exc) from exc

        try:
            setup = getattr(lib, "setup")
        except Exception as exc:
            del sys.modules[key]
            raise TypeError("There is no setup function inside your cog file!")

        try:
            setup(self)
        except Exception as exc:
            del sys.modules[key]
            raise ExtensionLoadError(key, exc) from exc
        else:
            self._extensions[key] = lib

    def add_extension(self, name: str, *, package: Optional[str] = None):
        name = self._resolver(name=name, package=package)
        if name in self._extensions.items():
            raise TypeError("Module is already loaded")

        spec = importlib.util.find_spec(name)
        if spec is None:
            raise TypeError(f"Extension {name} is not found!")

        self._extension_loader(spec, name)

    def remove_extension(self, name: str, *, package: Optional[str] = None):
        name = self._resolver(name=name, package=package)
        lib = self._extensions.get(name)
        if lib is None:
            raise TypeError("That Module isn't loaded")
        else:
            self.remove_cog(name)
            self._extensions.pop(lib)

    def create_modal(
        self,
        title: str,
        callback: dispatcher.Coro,
        components: list[ModalComponent],
        custom_id: int = utils.create_snowflake(),
    ):
        return Modal(self.state).create(
            title=title, callback=callback, components=components, custom_id=custom_id
        )

    async def _register_cog_commands(self, real: Cog):
        await asyncio.sleep(20)

        for command in real.guild_commands.values():
            for guild_id in command["guild_id"]:
                await self.application.register_guild_slash_command(
                    guild_id=guild_id,
                    name=command["name"],
                    description=command["description"],
                    callback=command["callback"],
                    options=command["options"],
                    default_permission=command["default_permission"],
                    cog=real,
                )
        for command in real.global_commands.values():
            await self.application.register_global_slash_command(
                name=command["name"],
                description=command["description"],
                callback=command["callback"],
                options=command["options"],
                default_permission=command["default_permission"],
                cog=real,
            )
